/*
 * GoatTracker mono-to-stereo converter
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "goattrk.h"

INSTR instr[MAX_INSTR];
unsigned char filtertable[256];
unsigned char wavetable[MAX_INSTR][MAX_WAVELEN*2];
unsigned char songorder[MAX_SONGS][6][MAX_SONGLEN+2];
unsigned char pattern[MAX_PATT][MAX_PATTROWS*3+3];
INSTR instrcopybuffer;
unsigned char wavecopybuffer[MAX_WAVELEN*2];
unsigned char songname[MAX_STR];
unsigned char authorname[MAX_STR];
unsigned char copyrightname[MAX_STR];
int pattlen[MAX_PATT];
int songlen[MAX_SONGS][6];

int dp;
int highestusedpattern = 0;
int oldorder = 3;
int neworder = 6;
unsigned char numsongs;
unsigned char numpatt;

int main(int argc, char **argv);
int loadsong(char *name);
int savesong(char *name);
int processsong(void);
void countpatternlengths(void);
void countdestpatternlengths(void);
void clearsong(int cs, int cp, int ci, int cn);

int main(int argc, char **argv)
{
  if (argc < 5)
  {
    printf("Usage: MONO2STR <sourcesong> <dest.song> <source chn> <dest chn>\n"
           "To convert mono to stereo: source chn = 3, dest chn = 6\n"
           "To convert stereo to mono: source chn = 6, dest chn = 3 (channels 1-3 saved)\n");

    return 1;
  }
  if (!strcmp(argv[1], argv[2]))
  {
    printf("ERROR: Source and destination are not allowed to be the same.");
    return 1;
  }

  sscanf(argv[3], "%u", &oldorder);
  sscanf(argv[4], "%u", &neworder);

  if (!loadsong(argv[1]))
  {
    printf("ERROR: Couldn't load source song.");
    return 1;
  }
  if (!savesong(argv[2]))
  {
    printf("ERROR: Couldn't save destination song.");
    return 1;
  }
  return 0;
}

int loadsong(char *name)
{
  char ident[4];
  int c;

  FILE *srchandle = fopen(name, "rb");
  if (srchandle)
  {
    fread(ident, 4, 1, srchandle);
    if (!memcmp(ident, "GTS!", 4))
    {
      int d;
      unsigned char length;
      int loadbytes;
      clearsong(1,1,1,1);

      /* Read infotexts */
      fread(songname, sizeof songname, 1, srchandle);
      fread(authorname, sizeof authorname, 1, srchandle);
      fread(copyrightname, sizeof copyrightname, 1, srchandle);

      /* Read songorderlists */
      fread(&numsongs, sizeof numsongs, 1, srchandle);
      for (d = 0; d < numsongs; d++)
      {
        for (c = 0; c < oldorder; c++)
        {
          fread(&length, sizeof length, 1, srchandle);
          loadbytes = length;
          loadbytes++;
          fread(songorder[d][c], loadbytes, 1, srchandle);
          /* Convert the old endmark */
          if (songorder[d][c][length] == 0xfe)
            songorder[d][c][length] = LOOPSONG;
        }
      }
      /* Read instruments */
      for (c = 1; c < MAX_INSTR; c++)
      {
        fread(&instr[c], sizeof(INSTR), 1, srchandle);
        loadbytes = instr[c].wavetableindex;
        instr[c].wavetableindex = 1;
        fread(wavetable[c], loadbytes, 1, srchandle);
      }
      /* Read patterns */
      fread(&numpatt, sizeof numpatt, 1, srchandle);
      for (c = 0; c < numpatt; c++)
      {
        fread(&length, sizeof length, 1, srchandle);
        fread(pattern[c], length, 1, srchandle);
      }
      /* Read filtertable */
      fread(filtertable, 256, 1, srchandle);

      countpatternlengths();
      fclose(srchandle);
      return 1;
    }
    fclose(srchandle);
  }
  return 0;
}

int savesong(char *name)
{
  char ident[] = {'G', 'T', 'S', '!'};
  FILE *handle;
  int c;

  handle = fopen(name, "wb");
  if (handle)
  {
    int d;
    unsigned char length;
    int writebytes;
    fwrite(ident, 4, 1, handle);

    /* Write infotexts */
    fwrite(songname, sizeof songname, 1, handle);
    fwrite(authorname, sizeof authorname, 1, handle);
    fwrite(copyrightname, sizeof copyrightname, 1, handle);

    fwrite(&numsongs, sizeof numsongs, 1, handle);
    /* Write songorderlists */
    for (d = 0; d < numsongs; d++)
    {
      for (c = 0; (c < oldorder) && (c < neworder); c++)
      {
        length = songlen[d][c]+1;
        fwrite(&length, sizeof length, 1, handle);
        writebytes = length;
        writebytes++;
        fwrite(songorder[d][c], writebytes, 1, handle);
      }
      for (c = oldorder; c < neworder; c++)
      {
      	length = 1;
      	unsigned char emptytrack[2] = {LOOPSONG, 0x00};
      	fwrite(&length, sizeof length, 1, handle);
      	writebytes = length;
      	writebytes++;
      	fwrite(emptytrack, writebytes, 1, handle);
      }
    }
    /* Write instruments */
    for (c = 1; c < MAX_INSTR; c++)
    {
      instr[c].wavetableindex = 2;
      for (d = 0; d < MAX_WAVELEN*2; d+= 2)
      {
        if (wavetable[c][d] == 0xff) break;
        instr[c].wavetableindex += 2;
      }
      fwrite(&instr[c], sizeof(INSTR), 1, handle);
      fwrite(wavetable[c], instr[c].wavetableindex, 1, handle);
      instr[c].wavetableindex = 1;
    }
    /* Write patterns */
    fwrite(&numpatt, sizeof numpatt, 1, handle);
    for (c = 0; c < numpatt; c++)
    {
      length = pattlen[c]*3+3;
      fwrite(&length, sizeof length, 1, handle);
      fwrite(pattern[c], length, 1, handle);
    }
    /* Write filtertable */
    fwrite(filtertable, 256, 1, handle);

    fclose(handle);
    return 1;
  }
  return 0;
}

void countpatternlengths(void)
{
  int c, d, e;

  highestusedpattern = 0;
  for (c = 0; c < MAX_PATT; c++)
  {
    for (d = 0; d <= MAX_PATTROWS; d++)
    {
      if (pattern[c][d*3] == ENDPATT) break;
    }
    pattlen[c] = d;
  }
  for (e = 0; e < MAX_SONGS; e++)
  {
    for (c = 0; c < 6; c++)
    {
      for (d = 0; d < MAX_SONGLEN; d++)
      {
        if (songorder[e][c][d] >= LOOPSONG) break;
        if (songorder[e][c][d] < MAX_PATT)
        {
          if (songorder[e][c][d] > highestusedpattern)
            highestusedpattern = songorder[e][c][d];
        }
      }
      songlen[e][c] = d;
    }
  }
}

void clearsong(int cs, int cp, int ci, int cn)
{
  int c;

  for (c = 0; c < 6; c++)
  {
    int d;
    if (cs)
    {
      for (d = 0; d < MAX_SONGS; d++)
      {
        memset(&songorder[d][c][0], 0, MAX_SONGLEN);
        if (!d)
        {
          songorder[d][c][0] = c;
          songorder[d][c][1] = LOOPSONG;
          songorder[d][c][2] = 0;
        }
        else
        {
          songorder[d][c][0] = LOOPSONG;
          songorder[d][c][1] = 0;
        }
      }
    }
  }
  if (cn)
  {
    memset(songname, 0, sizeof songname);
    memset(authorname, 0, sizeof authorname);
    memset(copyrightname, 0, sizeof copyrightname);
  }

  if (cp)
  {
    for (c = 0; c < MAX_PATT; c++)
    {
      int d;
      memset(&pattern[c][0], 0, MAX_PATTROWS*3);
      for (d = 0; d < MAX_PATTROWS; d++) pattern[c][d*3] = REST;
      for (d = MAX_PATTROWS; d <= MAX_PATTROWS; d++) pattern[c][d*3] = ENDPATT;
    }
  }
  if (ci)
  {
    for (c = 0; c < MAX_INSTR; c++)
    {
      memset(&instr[c], 0, sizeof(INSTR));
      memset(&wavetable[c][0], 0, MAX_WAVELEN*2);
      wavetable[c][2] = 0xff;
      wavetable[c][MAX_WAVELEN*2-2] = 0xff;
    }
    memset(&instrcopybuffer, 0, sizeof(INSTR));
    memset(wavecopybuffer, 0, MAX_WAVELEN*2);
  }
  countpatternlengths();
}


